/* iolocks.c
/* Copyright (c) 1988, Prime Computer, Inc., Natick, MA 01760
/*                     All Rights Reserved
*/

/* Lock package for IO benchmark.

   The locktab contains for each file a pointer to an array of locks
   for that file.  The actual size of a lock can be system dependent
   and is defined by iolock_t.  The following operations are defined
   on a lock pointed by p

         INIT_LOCK(p)        initialize the lock
         LOCK(p)             lock, spin until lock is available
         UNLOCK(p)           unlock
         CLOCK(p)            lock but don't spin if unavailable,
                             return 0 if lock was taken, 1 othewise
         IS_LOCKED(p)        return 1 if lock is taken, 0 if free

   Any implementation of locks that supports the above semantics is
   acceptable.

   The locktab is stored in shared memory that is allocated by call
   to get_shared_memory().  The locktab is initialized only when
   initlocks is called from controlling process.

   Variable server_locking non-zero indicates that locking is to be
   done by ioserver, in which case appropriate parameters are passed
   via the message mechanism.  server_locking zero means that locking
   is to be done in current process.

   Initial Version:     14 August 1988 by Barry Wolman

          Modified:     22 August 1988 by Barry Wolman
              1. Incorporate results of UNIX conversion
              2. Simplify lock semantics

          Modified:     11 September 1988 by Barry Wolman
              1. Don't print summary if file pointer is NULL

          Modified:     19 May 1989 by Barry Wolman
              1. Provide option to implement locks via lock manager process

          Modified:     18 May 1990 by Barry Wolman
              1. Cleaner control over no locking case
              2. Keep track of locking hits

          Modified:     19 July 1990 by Barry Wolman
              1. Don't get/put shared memory if locks_per_file = 0

          Modified:     24 July 1990 by Barry Wolman
              1. Use ERR macro instead of gripe

          Modified:      8 November 1990 by Barry Wolman
              1. Make lp static
*/

#include <stdio.h>
#include "ioincl.h"
#include "iolocks.h"    /* Defines iolock_t and locking macros */
#include "ioserver.h"

/* Following is the lock table;  it is actually stored in shared memory */

struct locktab {
    int         active[MAX_FILES],maxactive[MAX_FILES],number[MAX_FILES],
                 hits[MAX_FILES];
    iolock_t    file_locks[MAX_FILES];   /* Protects active, maxactive, number */
    iolock_t    *locks[MAX_FILES];
    iolock_t    first_lock;
    };

static struct locktab *lp = NULL;

iolock_t a_lock;

#define lockptr(fn,ln) ((iolock_t *)(lp -> locks[fn] + (ln)*(sizeof a_lock)))

char *get_shared_memory();

static int n_users,n_files,locks_per_file,server_locking;

/* Initialize locks for this process. Gloval variable user = 0
   means this is controlling process, user < 0 means this is
   ioserver process */

initlocks(nu,nf,lpf,s)
int nu,nf,lpf,s;
{
    register iolock_t *q;
    register int i,j;

    n_users = nu;
    n_files = nf;
    locks_per_file = lpf;
    server_locking = s;

#if UNIX
    if (server_locking) {

         if (user == 0) {

              /* Send parameters to lock server process */

              smsg.mitem[1] = n_users;
              smsg.mitem[2] = n_files;
              smsg.mitem[3] = locks_per_file;

              if (server_request(INIT_LOCK_MSG)) ERR("lock initialization failed");
              }
         }
    else {
#endif
         if (locks_per_file) {

              /* Allocate block of shared memory */

              lp = (struct locktab *)get_shared_memory(user);

              /* Only initialize if caller is iomain or ioserver */

              if (user <= 0) {
                   q = &(lp -> first_lock);
                   for (i = 0; i < n_files; i++) {
                        lp -> active[i] = lp -> maxactive[i] = lp -> number[i] =
                         lp -> hits[i] = 0;
                        INIT_LOCK(&(lp -> file_locks[i]));
                        lp -> locks[i] = q;
                        for (j = 0; j < locks_per_file; j++) {
                             INIT_LOCK(q);
                             q++;
                             }
                        }
                   }

              }
#if UNIX
         }
#endif
     }

/* Lock lock LN of file FN for user UN; sleep ST seconds if already locked */

lockrec(un,fn,ln,st)
int un,fn,ln,st;
    {
    if (locks_per_file == 0) return;   /* = 0 means no locking desired */

    if (un < 1 || un > n_users) {
         printf("Invalid user no %d\n",un);
         return;
         }

    if (fn < 0 || fn >= n_files) {
         printf("Invalid file no %d for user %d\n",fn,un);
         return;
         }

    if (ln < 0 || ln >= locks_per_file) {
         printf("Invalid lock no %d for user %d\n",ln,un);
         return;
         }

#if UNIX
    if (server_locking) {

         /* Request server to lock for us */

         smsg.mitem[1] = un;
         smsg.mitem[2] = fn;
         smsg.mitem[3] = ln;

         while(server_request(LOCK_MSG)) if (st) sleep(st);
         }
    else {
#endif
         /* Try to set the lock.  Use fast spin lock if no sleep
            time specified */

         if (st == 0) LOCK(lockptr(fn,ln));
         else while(CLOCK(lockptr(fn,ln)) == 0) {
             (lp -> hits[fn])++;
             sleep(st);
             }

         /* Increment count of locks in use and save if this is new high
            water mark. */

         LOCK(&(lp -> file_locks[fn]));

         if (++(lp -> active[fn]) > lp -> maxactive[fn])
              lp -> maxactive[fn] = lp -> active[fn];

         ++(lp -> number[fn]);

         UNLOCK(&(lp -> file_locks[fn]));
#if UNIX
         }
#endif
    }

/* This version of lockrec is provided for use by lock server process.
   Only conditional locking is done.  Failure to get BOTH locks causes
   failure return.  Any waiting is done by process that called lock
   server process */

server_lockrec(un,fn,ln)
int un,fn,ln;
    {
#if UNIX
    register
#endif
    iolock_t *p;

    p = lockptr(fn,ln);
    if (CLOCK(p) == 0) return(1);

    if (CLOCK(&(lp -> file_locks[fn])) == 0) {
         UNLOCK(p);
         return(1);
         }

    /* Increment count of locks in use and save if this is new high
       water mark. */

    if (++(lp -> active[fn]) > lp -> maxactive[fn])
         lp -> maxactive[fn] = lp -> active[fn];

    ++(lp -> number[fn]);

    UNLOCK(&(lp -> file_locks[fn]));

    return(0);
    }

/* Unlock lock LN of file FN for user UN */

unlockrec(un,fn,ln,st)
int un,fn,ln,st;
    {
    if (locks_per_file == 0) return;   /* = 0 means no locking desired */

    if (un < 1 || un > n_users) {
         printf("Invalid user no %d\n",un);
         return;
         }

    if (fn < 0 || fn >= n_files) {
         printf("Invalid file no %d for user %d\n",fn,un);
         return;
         }

    if (ln < 0 || ln >= locks_per_file) {
         printf("Invalid lock no %d for user %d\n",ln,un);
         return;
         }

#if UNIX
    if (server_locking) {

         /* Request server to unlock for us */

         smsg.mitem[1] = un;
         smsg.mitem[2] = fn;
         smsg.mitem[3] = ln;

         while(server_request(UNLOCK_MSG)) if (st) sleep(st);
         }
    else {
#endif
         if (*lockptr(fn,ln) == 0) {
              fprintf(stderr,"iouser %d: Lock %d for file %d should be taken but isn't\n",
               un,ln,fn);
              exit(1);
              }

         LOCK(&(lp -> file_locks[fn]));

         --(lp -> active[fn]);

         UNLOCK(&(lp -> file_locks[fn]));

         UNLOCK(lockptr(fn,ln));
#if UNIX
         }
#endif
    }

/* This version of unlockrec is provided for use by lock server process */

server_unlockrec(un,fn,ln)
int un,ln,fn;
    {
    if (*lockptr(fn,ln) == 0) {
         fprintf(stderr,"iouser %d: Lock %d for file %d should be taken but isn't\n",
          un,ln,fn);
         exit(1);
         }

   if (CLOCK(&(lp -> file_locks[fn])) == 0) return(1);

    --(lp -> active[fn]);

    UNLOCK(&(lp -> file_locks[fn]));

    UNLOCK(lockptr(fn,ln));

    return(0);
    }

/* Print lock statistics, return value is number of hits */

int statlocks(fp)
register FILE *fp;
    {
    register int i,j;
    int num,max,act[MAX_FILES],nhits;
    int errs = 0,tothits = 0;

    if (locks_per_file == 0) return(tothits);

    fprintf(fp,"\n FILE     LOCKS TAKEN     MAX ACTIVE    HITS\n");
    for(i = 0;i < n_files; i++) {
         lockinfo(i,&num,&max,&act[i],&nhits);
         tothits += nhits;
         fprintf(fp,"%5d   %13d       %8d%8d\n",i,num,max,nhits);
         }

    for(i = 0;i < n_files; i++) {
         if (act[i]) {
              fprintf(fp,"%d locks still held for file %d\n",act[i],i);
              errs++;
              }

         for(j = lockerrors(i,0); j >= 0; j = lockerrors(i,j+1)) {
              fprintf(fp,"Lock %d still locked for file %d\n",j,i);
              errs++;
              }
         }

    if (errs) printf("%d locking errors detected\n",errs);

    return(tothits);
    }

/* Return statistics for locks taken on file fn */

lockinfo(fn,np,map,ap,hp)
int fn,*np,*map,*ap,*hp;
    {
#if UNIX
    if (server_locking) {
         smsg.mitem[1] = fn;

         if (server_request(LOCK_INFO_MSG)) ERR("could not get lock info");

         *np = rmsg.mitem[1];
         *map = rmsg.mitem[2];
         *ap = rmsg.mitem[3];
         *hp = rmsg.mitem[4];
         }
    else {
#endif
         *np = lp -> number[fn];
         *map = lp -> maxactive[fn];
         *ap = lp -> active[fn];
         *hp = lp -> hits[fn];
#if UNIX
         }
#endif
    }

/* Check for errors on specified file starting at specified lock.
   Return number of lock still taken or -1 if no more errors for
   specified file */

lockerrors(fn,ln)
int fn,ln;
    {
    register int j,val;

#if UNIX
    if (server_locking) {
         smsg.mitem[1] = fn;
         smsg.mitem[2] = ln;

         if (server_request(LOCK_ERRORS_MSG)) ERR("could not check errors");

         val = rmsg.mitem[1];
         }
    else {
#endif
         val = -1;

         for (j = ln; j < locks_per_file; j++)
              if (IS_LOCKED(lockptr(fn,j))) {
                   val = j;
                   j = locks_per_file;
                   }
#if UNIX
         }
#endif
    return(val);
    }

/* Terminate all locks;  called by each background process and
   then by controlling user */

termlocks()
    {

#if UNIX
    if (server_locking) {
         if (user == 0) {
              if (server_request(TERM_LOCK_MSG)) ERR("Lock server could not terminate properly");
              }
         }
    else {
#endif
         /* Free shared memory */

         if (locks_per_file) free_shared_memory(user);
#if UNIX
         }
#endif
    }
