/* iouser.c
/* Copyright (c) 1988, Prime Computer, Inc., Natick, MA 01760
/*                     All Rights Reserved
*/

/* This is main program of the background task run as part of
   the IO benchmark.  It is initiated by controlling process
   and is passed the following arguments:

         1:   the user number
         2:   pathname of control file
         3:   pathname of syncfile
         4:   pathname of runfile
         5:   pathname of resultsfile
         6:   seed to pass to random number generator

   After intialization is done, we wait until syncfile appears;
   this is signal that timed portion of run should begin.  The
   file specified by runfile is created as soon as we start running
   and is deleted as the last thing we do.  All results are written
   to the results file.

   Initial Version: 12 August 1988 by Barry Wolman

          Modified: 29 August 1988 by Barry Wolman
              1. Add CP time component

          Modified: 10 September 1988 by Barry Wolman
              1. Close stdin and stdout to save on file descriptors
              2. Create runfile at start of run
              3. Close and reopen files if there aren't enough file
                 units to keep all files open
              4. Catch SIGTERM & SIGINT (UNIX) and LOGOUT$ (PRIMOS)
                 and do a cleanup
              5. Create runfile with 0666 mode

          Modified: 17 September 1988 by Barry Wolman
              1. Cleanup locks and exit if error

          Modified: 27 September 1988 by Barry Wolman
              1. Accumulate sum of squares of response times, per file
              2. Keep track of worse response time, per file

          Modified:  2 October 1988 by Barry Wolman
              1. Skip doing IO if calibrate flag set

          Modified: 15 February 1989 by Tom Olson
              1. Added transaction grouping read/write ratio and response
                 time distribution statistics.

          Modified: 22 February 1989 by Barry Wolman
              1. Call init_params
              2. Calculate distribution of transaction times
              3. Minor cleanup

          Modified: 25 May 1989 by Barry Wolman
              1. Add variants

          Modified: 24 August 1989 by Barry Wolman
              1. Don't attach to message queues if variant is zero
              2. Do all writes at end of loop instead of at start
              3. Conditionally compile server code

          Modified: 11 June 1990 by Barry Wolman
              1. Use setrandom instead of "srand"

          Modified: 20 June 1990 by Barry Wolman
              1. Use cptime instead of systime, rename variables
              2. Add code for counted run
              3. All durations in seconds
              4. Add checkpoints

          Modified: 17 July 1990 by Barry Wolman
              1. Add run validation code

          Modified: 19 July 1990 by Barry Wolman
              1. Store very latest time in validation data

          Modified: 24 July 1990 by Barry Wolman
              1. Reorder validation code to minimize window of vulnerability
              2. Use ERR macro instead of gripe
              3. Use runfile as logfile, create with 0600 mode and change
                 to 0666 at end of run, don't delete at end
              4. Keep log of all transactions, write to runfile

          Modified: 30 July 1990 by Barry Wolman
              1. Add setcleanup call

          Modified: 6 August 1990 by Barry Wolman
              1. Add error checks

          Modified: 10 August 1990 by Barry Wolman
              1. Fix definition of cleanup() to satisfy ANSI compilers

          Modified: 16 August 1990 by Barry Wolman
              1. If overall number of transactions isn't multiple of
                 number of users, distribute excess across users

          Modified: 7 November 1990 by Barry Wolman
              1. Take seed for iorandom() from arg list
              2. Don't complain if IORUN file can't be found during
                 cleanup initiated by a signal
*/

#include <stdio.h>
#include "ioincl.h"
#include "iodefaults.h"
#include "ioserver.h"

#if UNIX
#include <signal.h>
void cleanup();
#endif

#if PRIMOS
fortran void mkon$p();
fortran void error_handler();
fortran void cleanup();
int in_error_handler = 0;
#define time_t long
#endif

extern double random();

#define SECS_PER_DAY (24 * 60 * 60)

struct control control;

time_t time();
char *cvtime();
double time_of_day();

int the_file,the_lock,the_record,unit[MAX_FILES];
char runfile[168];

struct results results;

int status = 0;

main(argc,argv)
char **argv;
    {
    register int cycles,i,j;
    int start,finish,now,rfid,reads,terminate_run,tpu,basetpu,cpno;
    int time_interval,tpu_interval,checkno,docheckpoint;
    FILE *runfp;
    struct logrec log;
    struct rec rec;
    struct checkpoint *cp;
    double usert1,usert2,syst1,syst2,ts,tf,t,tranf,trans,trant;

    if (argc < 7) {
         fprintf(stderr,"iouser: Wrong number of arguments, %d\n",argc);
         return;
         }

    strcpy(runfile,argv[4]);

    setcleanup(cleanup);

#if UNIX
    fclose(stdin);
    fclose(stdout);
    signal(SIGTERM,cleanup);
    signal(SIGINT,cleanup);
#endif

#if PRIMOS
    mkon$p("LOGOUT$",7,cleanup);
    mkon$p("DISK_READ_ERR$",14,error_handler);
#endif

    user = atoi(argv[1]);
    setseed(atol(argv[6]));

    if (read_control("iouser",argv[2],&control)) return;

    /* Compute number of transactions to do if this is counted run */

    basetpu = tpu = control.trans / control.n_users;

    /* Distribute excess transactions */

    if (j = (control.trans - (tpu * control.n_users)))
         if (user <= j)
              tpu++;

#ifdef MESSAGEQUEUES
    if (control.variant) {

         /* Open message queues for communication with server */

         mypid = getpid();

         if ((smsgid = msgget(INKEY,0777)) < 0)
              ERR("msgget call failed for input queue");

         if ((rmsgid = msgget(OUTKEY,0777)) < 0)
              ERR("msgget call failed for output queue");
         }
#endif

    /* Create runfile that indicates we started running.  Mode of 0600
       means we haven't finished */

    if ((i = creat(runfile,0600)) < 0) ERRS("Could not create %s",runfile);

    if ((runfp = fdopen(i,"w")) == NULL) ERRS("Could not fdopen %s",runfile);

    /* Open as many files as we can, use negative unit values for
       those we can't open;  an attempt to use an unopened file will
       cause some other file to be closed */

    for(i = 0; i < control.n_files; i++) {
         results.opens[i] = results.closes[i] = results.uses[i]
          = results.freads[i] = results.fwrites[i] = 0;
         results.sum_time[i] = results.sumsq_time[i] = 0.0e0;
         results.worst_time[i] = -1.0e0;
         results.worst_rec[i] = -1;
         if ((unit[i] = open(control.path[i],2)) < 0) break;
         }

    for(i = i+1; i < control.n_files; i++) {
         results.opens[i] = results.closes[i] = results.uses[i]
          = results.freads[i] = results.fwrites[i] = 0;
         results.sum_time[i] = results.sumsq_time[i] = 0.0e0;
         results.worst_time[i] = -1.0e0;
         results.worst_rec[i] = -1;
         }

    results.tran_sum_time = results.tran_sumsq_time = results.tran_max_time = 0.0;

    initlocks(control.n_users,control.n_files,control.locks_per_file,
     control.variant & LOCKING_VARIANT);

    /* All initialization has been done, loop until control process
       creates the syncfile */

    while(!fexists(argv[3]));

    time(&start);
    finish = start + control.duration;
    cptimes(&usert1,&syst1);

    cp = &results.checkpoint[0];
    cp -> delta = 0;
    cp -> counter = 0;
    cp -> usert = cp -> syst = 0.0e0;

    checkno = 1;
    time_interval = control.duration / (NUMBER_OF_CHECKPOINTS+1);
    tpu_interval = basetpu / (NUMBER_OF_CHECKPOINTS+1);

    terminate_run = 0;
    for(cycles = 0; ; cycles++) {

        time(&now);

        /* Check termination */

        if (control.runtype == TIMED_RUN) {
             if (now >= finish) break;
             docheckpoint = (now-start) >= checkno * time_interval;
             }
        else {
             if (cycles >= tpu) break;
             docheckpoint = cycles >= checkno * tpu_interval;
             }

        /* If necessary, do the checkpoint */

        if (docheckpoint && checkno <= NUMBER_OF_CHECKPOINTS) {
             cp = &results.checkpoint[checkno];
             cp -> counter = cycles;
             cp -> delta = now-start;
             cptimes(&usert2,&syst2);
             cp -> usert = usert2-usert1;
             cp -> syst = syst2-syst1;
             checkno++;
             }

        /* Do the IO required for a transaction */

         trans=time_of_day();

         /* Following loop will do R-W reads followed by W reads & writes.
            CP time to be associated with the transaction is consumed
            in separate chunks, rather than all at once */

         for(reads = 0; reads < control.n_reads; reads++) {

              /* Pick a file and record at random; allow for the case where
                 random returns exactly 1.0 */

              if ((the_file = control.n_files * random()) >= control.n_files)
                   the_file=0;

              /* If file we selected isn't open, terminate run if we can't
                 reopen it */

              if (unit[the_file] < 0)
                   if(reopen(the_file)) {
                        terminate_run++;
                        break;
                        }

              if ((the_record = control.n_recs * random()) > control.n_recs - 1)
                   the_record=0;

              the_lock = (the_record * control.locks_per_file) / control.n_recs;

              /* Return from following lock call occurs when
                 we've got the lock */

              lockrec(user,the_file,the_lock,control.lock_sleep_time);

              if (control.calibrate) ts = tf = 0.0;
              else {
                   ts = time_of_day();

                   getrec(unit[the_file],the_record,&rec,control.reclen);

                   results.reads++;
                   results.freads[the_file]++;
                   tf = time_of_day();
                   }

              /* If requested, burn some CP time */

              if (control.param) work(control.param/control.n_reads);

              if (!control.calibrate){

                   if (reads >= (control.n_reads - control.n_writes)) {

                        /* Rewrite this record */

                        rec.counter = cycles;
                        rec.user = user;

                        log.l_file = the_file;
                        log.l_record = the_record;
                        log.l_counter = cycles;

/**********************************************************************

Note: The following comment applies if iobench is built with REALLOCKING
undefined or if LOCKS=0 is used.  If iobench is built and used with
locking, the following comment is not relevant.

The following two lines of code (time & putrec) really should be
executed with interrupts disabled or as a critical region.  Suppose this
process (say, user I) is executing at time TI=120 seconds.  If user I
stops running after the 'time' call and before the 'putrec' call, there
is a small, but finite chance that another iouser process (say, user J)
that runs before user I is restarted, e.g.  at time TJ=122 seconds,
might select the same record we're trying to write and record its own
validation time of TJ=122 seconds.  When user I is restarted, it will
overwrite the time written by user J with the now out of date time of
TI=120 seconds.  When iobench later tries to validate the run and checks
the data for user J, it will expect to find a time of 122 seconds, but
will actually find time 120 seconds; this difference will cause a
validation error unless the validation code is smart enough to detect
and compensate for this situation.

In order to avoid the validation errors that might occur with this
window open, iouser keeps a complete log of all records written.  The
case described above can be detected and the validation error avoided.

Note: the above comments apply only when REALLOCKING is NOT specified or
when LOCKS=0 is used.  If iobench is built with real locking and locking
is used during the run, no other user process can be accessing this
record and the window of vulnerability isn't open.

************************Start Critical Region ************************/

                        time(&rec.time);
                        putrec(unit[the_file],the_record,&rec,control.reclen);

/********************** End Critical Region **************************/

                        log.l_time = rec.time;

                        if (fwrite(&log,sizeof log,1,runfp) != 1)
			 ERR("Could not log a file access");

                        results.writes++;
                        results.fwrites[the_file]++;
                        }
                   }

              unlockrec(user,the_file,the_lock,control.lock_sleep_time);

              /* Update statistics for file.  Be careful to check for case
                 where run spans midnight */

              if ((t = tf-ts) < 0) t += (double)SECS_PER_DAY;

              results.sum_time[the_file] += t;
              results.sumsq_time[the_file] += (t*t);

              if (t > results.worst_time[the_file]) {
                   results.worst_time[the_file] = t;
                   results.worst_rec[the_file] = the_record;
                   }

              results.uses[the_file]++;
             }

         if (terminate_run) break;

         /* Update statistics for transaction.  Be careful to check
            for case where run spans midnight */

         tranf=time_of_day();
         if ((trant = tranf-trans) < 0) trant += (double)SECS_PER_DAY;

         /* Update transaction response time distribution table.  If
            transaction response > 24 seconds put in 25th entry of
            table */

         j = (int)trant;
         if (j > 24) j = 25;
         results.resp_dist[j]++;

         results.tran_sum_time += trant;
         results.tran_sumsq_time += (trant * trant);
         if (trant > results.tran_max_time) results.tran_max_time = trant;
         }

    /* Finished running, write out results into record user
       of results file */

    time(&finish);
    cptimes(&usert2,&syst2);
    results.userno = user;
    results.counter = cycles;
    results.start = start;
    results.finish = finish;
    results.usert = usert2-usert1;
    results.syst = syst2-syst1;

    while((rfid = open(argv[5],1)) < 0) sleep(1);

    putrec(rfid,user,&results,sizeof(results));
    if (close(rfid) < 0) ERRS("Could not close %s",argv[5]);
    if (fclose(runfp)) ERRS("Could not close %s",runfile);

#ifdef IODEBUG
    fprintf(stderr,"Started  %s\n",cvtime(start,0));
    fprintf(stderr,"     %8d cycles\n",cycles);
    fprintf(stderr,"     %8.3f cpu seconds\n",results.cpu);
    fprintf(stderr,"     %8.3f io  seconds\n",results.io);
    fprintf(stderr,"Finished %s\n",cvtime(finish,0));
#endif
    cleanup(0);
    }

#if PRIMOS
fortran
#endif
void cleanup(signo)
int signo;
    {
    register int i;

    if (signo) status = signo;

    for(i = 0; i < control.n_files; i++) close(unit[i]);

    termlocks();

    /* Set access to runfile to 0666 as last thing we do.  Don't complain
       if we can't set mode unless this is normal exit */

    if ((chmod(runfile,0666) < 0) & (signo == 0))
         ERRS("Could not change mode of %s",runfile);

    exit(status);
    }

/* Reopen specified file */

reopen(fn)
register int fn;
    {
    register int k;

    /* Look after and then before current file for file to close */

    for(k = fn+1; k != fn; k++) {
         if (k == control.n_files) k = 0;

         if (unit[k] >= 0) {
              if (close(unit[k]) < 0) ERRS("Could not close %s",control.path[k]);
              unit[k] = -1;
              results.closes[k]++;

              if ((unit[fn] = open(control.path[fn],2)) < 0) {
                   MSGS("Could not reopen %s",control.path[fn]);
                   return(1);
                   }

              results.opens[fn]++;
              return(0);
              }
         }

    /* Couldn't find a file to close (improbable!) */

    MSG("Could not find at file to close");
    return(1);
    }

#if PRIMOS
fortran void error_handler()
    {
    printf("DISK READ ERROR occured for file %d, lock %d\n",the_file,the_lock);
    if (in_error_handler) exit(1);
    status = 1;
    in_error_handler++;
    unlockrec(user,the_file,the_lock,control.lock_sleep_time);
    cleanup(0);
    }
#endif
